from typing import Dict, List, Tuple

from spb_label import sdk as spb_label

from spb_apps.curate.superb_curate import SuperbCurate
from spb_apps.label.superb_label import SuperbLabel
from spb_apps.utils.converter import convert_yolo_bbox
from spb_apps.utils.utils import read_info_from_zip_yolo


class SuperbApps:
    """
    A class to manage and interact with Superb applications Label and Curate.
    """

    def __init__(
        self,
        team_name: str,
        access_key: str,
        platform: str,
        project_id: str = "",
        project_name: str = "",
        dataset_name="",
    ):
        """
        Initializes the SuperbApps instance with necessary details and clients based on the platform.

        Args:
            team_name (str): The name of the team.
            access_key (str): The access key for authentication.
            platform (list): A list of platforms to initialize (e.g., ['label', 'curate']).
            project_id (str, optional): The ID of the project for labeling. Defaults to "".
            dataset_name (str, optional): The name of the dataset for Curate. Defaults to "".
        """
        self.team_name = team_name
        self.access_key = access_key
        if platform == "label":
            self.client = SuperbLabel(
                team_name=team_name,
                access_key=access_key,
                project_id=project_id,
                project_name=project_name,
            )

            self.label_project_name = self.client.client._project.name
        if platform == "curate":
            self.client = SuperbCurate(
                team_name=team_name,
                access_key=access_key,
                dataset_name=dataset_name,
            )
        self.platform = platform

    def download_image_by_key(self, data_key: str, path: str = None):
        """
        Downloads an image by its data key.

        Args:
            data_key (str): The data key of the image to download.
            path (str, optional): The local file path where the image will be saved. Defaults to None.
        """
        if self.platform == "label":
            _, label = self.client.get_labels(data_key=data_key)
            self.client.download_image(label=label[0], path=path)
        if self.platform == "curate":
            self.client.download_image(data_key=data_key, download_path=path)

    def download_image_by_filter(
        self,
        tags: list = [],
        data_key: str = "",
        status: list = [],
        path: str = None,
    ):
        """
        Downloads images by applying filters such as tags, data key, and status.

        Args:
            tags (list, optional): A list of tags to filter images. Defaults to [].
            data_key (str, optional): A data key to filter images. Defaults to "".
            status (list, optional): A list of statuses to filter images. Defaults to [].
            path (str, optional): The local file path where the images will be saved. Defaults to None.
        """
        from concurrent.futures import ThreadPoolExecutor

        if self.platform == "label":

            def download(label):
                self.client.download_image(label=label, path=path)

            count, labels = self.client.get_labels(
                tags=tags, data_key=data_key, status=status
            )
            print(f"Downloading {count} data to {path}")
            if count > 50:
                with ThreadPoolExecutor(max_workers=4) as executor:
                    executor.map(download, labels)
            else:
                for label in labels:
                    download(label)
        else:
            print("Curate does not have filters")

    def get_width_height(
        self, data_hanler: spb_label.DataHandle = None, data_key: str = ""
    ) -> Tuple[int, int]:
        """
        Retrieves the width and height of an image based on its data key.

        This method supports both 'label' and 'curate' platforms. It fetches the image's dimensions
        by utilizing the respective client's method to get the width and height.

        Args:
            data_key (str): The unique identifier for the image.

        Returns:
            Tuple[int, int]: A tuple containing the width and height of the image.
        """
        if self.platform == "label":
            if data_hanler == None:
                print("Label takes data handler object from spb_label")
                return
            return self.client.get_width_height(label=data_hanler)
        if self.platform == "curate":
            if data_key == "":
                print("Curate takes data key")
                return
            return self.client.get_width_height(data_key=data_key)

    def get_labels(
        self,
        data_key: str = "",
        tags: list = [],
        assignees: list = [],
        status: list = [],
        all: bool = False,
    ) -> Tuple[int, List]:
        """
        Retrieves labels based on filters or all labels if specified.

        Args:
            data_key (str, optional): A data key to filter labels. Defaults to "".
            tags (list, optional): A list of tags to filter labels. Defaults to [].
            assignees (list, optional): A list of assignees to filter labels. Defaults to [].
            status (list, optional): A list of statuses to filter labels. Defaults to [].
            all (bool, optional): If True, retrieves all labels ignoring other filters. Defaults to False.

        Returns:
            Tuple[int, List]: A tuple containing the count of labels and a list of labels.
        """
        count, labels = self.label_client.get_labels(
            data_key=data_key,
            tags=tags,
            assignees=assignees,
            status=status,
            all=all,
        )

        return count, labels

    def change_project(self, project_name: str):
        """
        Changes the project context for the label client.

        Args:
            project_name (str): The name of the project to switch to.
        """
        self.label_client.client.set_project(name=project_name)
        self.label_project_name = self.label_client.client._project.name

    def get_label_interface(self) -> Dict:
        """
        Retrieves the label interface configuration.

        Returns:
            Dict: The label interface configuration.
        """
        lb_interface = self.label_client.client.project.label_interface
        return lb_interface

    def make_bbox(self, class_name: str, annotation: list, data_key: str = ""):
        """
        Creates a bounding box based on the specified platform ('label' or 'curate').

        Args:
            class_name (str): The class name associated with the bounding box.
            annotation (list): A list containing the x, y coordinates, width, and height of the bounding box.
            data_key (str, optional): The unique identifier for the image. Required for 'curate' platform.

        Returns:
            The result of the bounding box setting operation, which varies by platform.
            For 'label', it returns the result of `client.bbox_setting` with class_name and annotation.
            For 'curate', it additionally requires a data_key and returns the result of `client.bbox_setting` with data_key, class_name, and annotation.
            If the required data_key is not provided for 'curate', it prints a message and returns None.
        """
        if self.platform == "label":
            return self.client.bbox_setting(
                class_name=class_name, annotation=annotation
            )
        if self.platform == "curate":
            if data_key == "":
                print("To make a Curate bbox, you must provide a data_key.")
                return
            return self.client.bbox_setting(
                data_key=data_key, class_name=class_name, annotation=annotation
            )

    def upload_images(self, images_path: str, dataset_name: str = ""):
        """
        Uploads images to the specified platform ('label' or 'curate').

        For the 'label' platform, a dataset name must be specified. If not, an error message is printed.
        For the 'curate' platform, images are uploaded without needing a dataset name.

        Args:
            images_path (str): The path to the images to be uploaded.
            dataset_name (str, optional): The name of the dataset to upload the images to. Required for 'label' platform.

        Returns:
            The result of the image upload operation, which varies by platform.
            For 'label', it returns the result of `client.upload_image` with images_path and dataset_name.
            For 'curate', it returns the result of `client.curate_upload_images` with images_path.
            If the required dataset_name is not provided for 'label', it prints a message and returns None.
        """
        if self.platform == "label":
            if dataset_name == "":
                print("Must specify a dataset name when uploading to Label")
                return
            return self.client.upload_image(images_path, dataset_name)
        if self.platform == "curate":
            return self.client.curate_upload_images(images_path)

    def upload_annotations(
        self, data_key: str, annotations: list, format: str, classes: list = ""
    ):
        """
        Uploads annotations to the specified platform ('label' or 'curate') based on the provided format.

        For the 'label' platform and 'yolo' format, it converts YOLO annotations to the platform's format and uploads them.
        Requires a list of classes for 'yolo' format. If classes are not provided, it prints a message and returns None.

        Args:
            data_key (str): The unique identifier for the data to which the annotations belong.
            annotations (list): A list of annotations to be uploaded.
            format (str): The format of the annotations ('yolo', etc.).
            classes (list, optional): The list of classes associated with the annotations. Required for 'yolo' format.

        Returns:
            None. Prints a message if classes are not provided for 'yolo' format.
        """
        if self.platform == "label":
            _, label = self.client.get_labels(data_key=data_key)
            data_handler = label[0]
            if format == "yolo":
                if classes == "":
                    print(
                        "To upload yolo annotations, you must provide a classes."
                    )
                    return
                width, height = self.get_width_height(data_handler)
                converted_annotations = convert_yolo_bbox(
                    data_key, annotations, classes, width, height
                )
                if converted_annotations is not None:
                    self.client.upload_annotation(
                        data_handler, converted_annotations
                    )

    def read_yolo_zip(self, dataset_file: str):
        """
        Reads information from a YOLO formatted zip file.

        Args:
            dataset_file (str): The path to the YOLO formatted zip file.

        Returns:
            The information read from the zip file.
        """
        return read_info_from_zip_yolo(dataset_file)



import json
import os
from pathlib import Path
from zipfile import BadZipFile, ZipFile

from directory_tree import display_tree


def separate_batches(image_batch_size: int, to_batch_list: list):
    image_paths = []
    number_of_iteration = len(to_batch_list) // image_batch_size + 1
    for i in range(number_of_iteration):
        start = i * image_batch_size
        end = (
            (i + 1) * image_batch_size
            if (i + 1) * image_batch_size < len(to_batch_list)
            else False
        )
        if end is False:
            subset_image_paths = to_batch_list[start:]
        else:
            subset_image_paths = to_batch_list[start:end]
        image_paths.append(subset_image_paths)

    return image_paths


def read_info_from_zip_yolo(dataset_file: str):
    try:
        second_level_folders = set()
        expected_second_level_folders = ["images", "labels"]
        image_extensions = ["jpg", "jpeg", "png", "bmp"]
        classes_txt = None

        with ZipFile(dataset_file, "r") as zip_file:
            name_list = zip_file.namelist()

            file_list = [
                f
                for f in name_list
                if "__MACOSX" not in f and "DS_Store" not in f
            ]

            top_level_folder = os.path.commonprefix(file_list)

            if top_level_folder:  # top_level/images/0001.jpg
                for item in file_list:
                    if item.count("/") == 2 and item.endswith("/"):
                        second_level_folders.add(item.split("/")[1])

                    elif item.count("/") == 1 and item.endswith("classes.txt"):
                        classes_txt = item

                image_path_list = [
                    f
                    for f in file_list
                    if f.split("/")[-1].split(".")[-1].lower()  # check ext
                    in image_extensions
                    and f.split("/")[-2] == expected_second_level_folders[0]
                ]

                annotation_path_list = [
                    f
                    for f in file_list
                    if f.split("/")[-2] == expected_second_level_folders[1]
                    and f.endswith(".txt")
                ]

            else:  # images/0001.jpg
                for item in file_list:
                    if item.count("/") == 1:
                        second_level_folders.add(item.split("/")[0])
                    elif item.count("/") == 0 and item.endswith("classes.txt"):
                        classes_txt = item

                image_path_list = [
                    f
                    for f in file_list
                    if f.split("/")[-1].split(".")[-1].lower()
                    in image_extensions
                    and f.split("/")[-2] == expected_second_level_folders[0]
                ]

                annotation_path_list = [
                    f
                    for f in file_list
                    if "/" in f
                    and f.endswith(".txt")
                    and f.split("/")[-2] == expected_second_level_folders[1]
                ]

            if set(second_level_folders) != set(expected_second_level_folders):
                if len(second_level_folders) == 0:
                    raise Exception(
                        "Required folders, 'images' and 'labels' are missing in your zip file."
                    )
                # images 폴더만 있는 경우
                elif (
                    expected_second_level_folders[0] in second_level_folders
                    and expected_second_level_folders[1]
                    not in second_level_folders
                ):
                    raise Exception(
                        "Required folder, 'labels', is missing in your zip file."
                        f"Your zip file includes {str(second_level_folders)[1:-1]}."
                    )
                # labels 폴더만 있는 경우
                elif (
                    expected_second_level_folders[0]
                    not in second_level_folders
                    and expected_second_level_folders[1]
                    in second_level_folders
                ):
                    raise Exception(
                        "Required folder, 'images', is missing in your zip file."
                        f"Your zip file includes {str(second_level_folders)[1:-1]}."
                    )

            if not classes_txt:
                raise Exception(
                    "Required file, 'classes.txt' is missing in your zip file."
                )

            if len(image_path_list) == 0:
                raise Exception(
                    "The 'images' folder contains no image files with valid formats (PNG, JPEG, JPG, BMP)."
                )

            if len(annotation_path_list) == 0:
                raise Exception(
                    "The 'labels' folder contains no label files with valid formats (txt)."
                )

            zip_file.extractall(path=Path(dataset_file).parent)

        return image_path_list, annotation_path_list, classes_txt

    except BadZipFile as e:
        file_name = dataset_file.split("/")[-1]
        raise Exception(f"[Invalid zip file] '{file_name}' {e}.")

    except Exception as e:
        print("[Zip File]")
        with ZipFile(dataset_file, "r") as zip_file:
            zip_file.extractall(path=Path(dataset_file).parent)
            display_tree(str(Path(dataset_file).parent), max_depth=2)
        print("")
        raise Exception(f"[Invalid zip file] {e}")



def convert_yolo_bbox(data_key: str, annotations: list, classes: list,  width:int, height: int):
    converted_annotations = []
    for anno in annotations:
        class_number, x_norm, y_norm, w_norm, h_norm = anno[0], anno[1], anno[2], anno[3], anno[4]
        class_name = classes[class_number]
        # Check normalization
        if any(value > 1 for value in (x_norm, y_norm, w_norm, h_norm)):
            print(
                f"[WARNING] [Invalid zip file] {data_key}.txt\n"
                + f"[[Class index] [x_center] [y_center] [width] [height]]: [{int(class_number)} {x_norm} {y_norm} {w_norm} {h_norm}]\n"
                + f" YOLO format coordinates, width, and height must be normalized (0 - 1)."
            )
            return None
    
        w = w_norm * width
        h = h_norm * height
        x = x_norm * width - w / 2
        y = y_norm * height - h / 2
        converted_annotations.append([class_name, [x, y, w, h]])
    
    return converted_annotations


import os
from io import BytesIO
from pathlib import Path
from typing import List, Optional, Tuple
from uuid import uuid4

import phy_credit
import requests
from PIL import Image
from spb_label import sdk as spb_label
from spb_label.exceptions import ParameterException
from spb_label.utils import SearchFilter


class SuperbLabel(object):
    def __init__(
        self,
        team_name: str,
        access_key: str,
        project_id: str = "",
        project_name: str = "",
    ):
        """
        Initializes the SuperbLabel class with necessary details for operation.

        :param team_name: The name of the team.
        :param access_key: The access key for authentication.
        :param project_id: The ID of the project to be set for the client.
        """
        self.team_name = (
            team_name  # Assign the provided team name to an instance variable.
        )
        self.client = spb_label.Client(
            team_name=team_name,
            access_key=access_key,
            project_id=project_id if project_id else None,
            project_name=project_name if project_name else None,
        )

    def download_export(
        self,
        input_path: str,
        export_id: str,
    ):
        """
        Downloads an export from the server to a local path.

        :param input_path: The local file path where the export will be saved.
        :param export_id: The ID of the export to download.
        """
        print("[INFO] Checking for the export to be downloaded...")
        download_url = self.client.get_export(
            id=export_id
        ).download_url  # Retrieve the download URL for the export.
        r = requests.get(
            download_url
        )  # Make a GET request to the download URL.
        if r.status_code == 200:  # Check if the request was successful.
            print("Saving export to local path")
            Path(input_path).parents[0].mkdir(
                parents=True, exist_ok=True
            )  # Ensure the directory exists.
            with open(
                input_path, "wb"
            ) as f:  # Open the file in write-binary mode.
                f.write(
                    r.content
                )  # Write the content of the response to the file.
        else:
            print(
                f"Failed to download the file. Status code: {r.status_code}"
            )  # Log failure.

    def get_labels(
        self,
        data_key: str = "",
        tags: list = [],
        assignees: list = [],
        status: list = [],
        all: bool = False,
    ) -> Tuple[int, List]:
        """
        Retrieves labels based on provided filters or all labels if specified.

        :param data_key: Filter for specific data key.
        :param tags: Filter for specific tags.
        :param assignees: Filter for specific assignees.
        :param status: Filter for specific status.
        :param all: If True, ignores other filters and retrieves all labels.
        :return: A tuple containing the count of labels and a list of labels.
        """
        if all:  # If retrieving all labels.
            next_cursor = None
            count, labels, next_cursor = self.client.get_labels(
                cursor=next_cursor
            )  # Retrieve all labels.
        else:  # If using filters.
            filter = SearchFilter(
                project=self.client.project
            )  # Create a search filter.
            if data_key:
                filter.data_key_matches = data_key  # Set data key filter.
            if tags:
                filter.tag_name_all = tags  # Set tags filter.
            if assignees:
                filter.assignee_is_any_one_of = (
                    assignees  # Set assignees filter.
                )
            if status:
                filter.status_is_any_one_of = status  # Set status filter.
            next_cursor = None
            count, labels, next_cursor = self.client.get_labels(
                filter=filter, cursor=next_cursor
            )  # Retrieve filtered labels.

        return count, labels  # Return the count and list of labels.

    def download_image(self, label: spb_label.DataHandle, path: str = None):
        """
        Downloads an image associated with a label to a specified path.

        :param label: The label data handle containing the image to download.
        :param path: The local file path where the image will be saved. If None, defaults to the label's default path.
        """
        label.download_image(
            download_to=path
        )  # Download the image to the specified path.

    def get_width_height(self, label: spb_label.DataHandle) -> Tuple[int, int]:
        """
        Downloads an image associated with a label to a specified path, returns its width and height, and deletes the image.

        :param label: The label data handle containing the image to download.
        :return: A tuple containing the width and height of the downloaded image.
        """
        image_url = label.get_image_url()
        response = requests.get(image_url)
        img = Image.open(BytesIO(response.content))
        width, height = img.size

        return width, height

    def bbox_setting(self, class_name: str, annotation: list):
        """
        Creates a bounding box setting for a given class name and annotation coordinates.

        :param class_name: The class name associated with the bounding box.
        :param annotation: A list containing the coordinates of the bounding box in the order [x, y, width, height].
        :return: A tuple containing the class name and a dictionary with the bounding box coordinates.
        """
        bbox = {
            "class_name": class_name,
            "annotation": {
                "coord": {
                    "x": annotation[0],
                    "y": annotation[1],
                    "width": annotation[2],
                    "height": annotation[3],
                }
            },
        }

        return bbox

    def upload_image(self, image_paths: list, dataset_name: str):
        for path in image_paths:
            try:
                self.client.upload_image(path, dataset_name)
            except ParameterException as e:
                print(f"[ERROR]: Uploading went wrong: {e}")

        return

    def upload_annotation(
        self, label: spb_label.DataHandle, annotations: list
    ):
        for anno in annotations:
            bbox = self.bbox_setting(anno[0], anno[1])
            label.add_object_label(bbox)
        label.update_info()

    def build_label_interface_from_yolo(
        self,
        dataset_file: str,
        class_file_path: str,
    ):
        existing_label_interface = self.client.label_interface
        if existing_label_interface:
            label_interface = phy_credit.imageV2.LabelInterface.from_dict(
                existing_label_interface
            )
            object_detection = phy_credit.imageV2.ObjectDetectionDef.from_dict(
                existing_label_interface.get("object_detection")
            )
        else:
            # get default label_interface and object_detection (json)
            label_interface = phy_credit.imageV2.LabelInterface.get_default()
            object_detection = (
                phy_credit.imageV2.ObjectDetectionDef.get_default()
            )

        # create class map from classes.txt and create label interface for project
        category_map = dict()
        with open(
            os.path.join(Path(dataset_file).parent, class_file_path), "r"
        ) as f:
            lines = f.readlines()

        for i, line in enumerate(lines):
            category_map[i] = line.rstrip("\n")
            bbox_suite_class_id = str(uuid4())
            bbox_suite_class_name = category_map[i]

            object_detection.add_box(
                name=bbox_suite_class_name, id=bbox_suite_class_id
            )

        label_interface.set_object_detection(object_detection=object_detection)
        print("[Classes]")
        for value in category_map.values():
            print(value, end=",")
        print("")
        print("")
        return label_interface, category_map



import json
import os
import time
from pathlib import Path
from timeit import default_timer as timer
from typing import List, Tuple
from uuid import uuid4

import requests
import spb_curate
from spb_curate.error import ConflictError

from spb_apps.utils.utils import separate_batches

SLEEP_INTERVAL = 5  # time in seconds to wait between loop iterations


class SuperbCurate(object):
    def __init__(
        self,
        team_name: str,
        access_key: str,
        dataset_name: str = "",
        slice_name: str = "",
        is_dev: bool = False,
    ):
        """
        Initializes the SuperbCurate class with team, dataset, and slice details.
        Optionally sets the environment to development mode.

        :param team_name: Name of the team
        :param access_key: Access key for authentication
        :param dataset_name: Name of the dataset (optional)
        :param slice_name: Name of the slice within the dataset (optional)
        :param is_dev: Flag to set the environment to development mode (optional)
        """
        self.team_name: str = team_name  # Team name
        self.access_key: str = access_key  # Access key for authentication
        self.dataset_name: str = dataset_name  # Dataset name (optional)
        # Set global variables in spb_curate module for team name and access key
        spb_curate.team_name = self.team_name
        spb_curate.access_key = self.access_key
        # Set API base URL to development environment if in development mode
        if is_dev:
            spb_curate.api_base = "https://api.dev.superb-ai.com"

        # Attempt to fetch the dataset if a dataset name is provided
        if len(dataset_name) > 0:
            try:
                self.dataset = spb_curate.fetch_dataset(name=dataset_name)
            except ConflictError:
                # Notify the user if the dataset does not exist and needs to be created
                print(
                    f"Dataset does not exist, Creating Dataset {dataset_name}"
                )
                self.dataset = spb_curate.create_dataset(
                    name=dataset_name, description="Demo dataset."
                )
        # Attempt to fetch the slice from the dataset if a slice name is provided
        if len(slice_name) > 0:
            self.slice = self.dataset.fetch_slice(
                name=slice_name
            )  # Fetch the slice

    def curate_prep_images(self, images_path: list) -> List[spb_curate.Image]:
        """
        This will prep local images for Superb Curate by creating a list of
        Superb Curate images.

        Args:
            images_path (list): list of paths to upload

        Returns:
            List[spb_curate.Image]: list of curate images to upload
        """
        curate_images: List[spb_curate.Image] = []
        for image in images_path:
            curate_images.append(
                spb_curate.Image(
                    key=image.split("/")[-1],
                    source=spb_curate.ImageSourceLocal(asset=image),
                    metadata={},
                )
            )

        return curate_images

    def curate_upload_images(self, image_path: list):
        """
        Uploads images in batches to the dataset.

        Args:
            image_path (list): List of image paths to upload.
        """
        # Separate images into batches of 500
        seperated_images = separate_batches(
            image_batch_size=500, to_batch_list=image_path
        )
        total_loops = len(seperated_images)  # Total number of batches
        for idx, sep_images in enumerate(seperated_images, start=1):
            # Prepare images for upload
            curate_images = self.curate_prep_images(images_path=sep_images)
            # Create bulk image import job
            image_import_job = spb_curate.Image.create_bulk(
                dataset_id=self.curate_dataset["id"], images=curate_images
            )

            print(f"created an image import job: ({idx}/{total_loops})")
            start_time = timer()  # Start timing the upload process

            while True:
                # Fetch the current status of the import job
                image_import_job = spb_curate.Job.fetch(
                    id=image_import_job["id"]
                )
                print(
                    f"job progress: {image_import_job['progress']}", end="\r"
                )

                # Check if the job is complete
                if image_import_job["status"] == "COMPLETE":
                    # Print any failure details if present
                    if image_import_job["result"]["fail_detail"]:
                        print(image_import_job["result"]["fail_detail"])
                        print(image_import_job["fail_reason"])
                    break  # Exit the loop if job is complete

                if image_import_job["status"] == "FAILED":
                    if image_import_job["result"]["fail_detail"]:
                        print(
                            "[INFO] Fail detail: ",
                            image_import_job["result"]["fail_detail"],
                        )
                        print(
                            "[INFO] Fail reason: ",
                            image_import_job["fail_reason"],
                        )
                    break
                time.sleep(
                    SLEEP_INTERVAL
                )  # Wait before checking the status again

            print(
                f"total time: {timer() - start_time}"
            )  # Print total upload time

    def curate_prep_annotations(self, annotation: list) -> List:
        """
        Prepares annotations for upload by creating a list of spb_curate.Annotation objects.

        Args:
            annotation (list): List of dictionaries containing annotation details.

        Returns:
            List[spb_curate.Annotation]: List of prepared annotations for upload.
        """
        curate_annotations: List[spb_curate.Annotation] = []
        for anno in annotation:
            meta = anno.get(
                "metadata", {"iscrowd": 0}
            )  # Safely get metadata or default
            curate_annotations.append(
                spb_curate.Annotation(
                    image_key=anno[
                        "data_key"
                    ],  # Key of the image being annotated
                    annotation_class=anno[
                        "class_name"
                    ],  # Class of the annotation
                    annotation_type=anno[
                        "annotation_type"
                    ],  # Type of annotation (e.g., bounding box, segmentation)
                    annotation_value=anno[
                        "annotation"
                    ],  # The actual annotation data
                    metadata=meta,
                )
            )

        return curate_annotations

    def curate_upload_annotations(
        self,
        annotation_list: list,
    ):
        """
        Uploads annotations in batches to the dataset.

        Args:
            annotation_list (list): List of annotations to upload. Format is a
                                    list of dicts that are made up of "class_name", "annotation", "data_key", "annotation_type", "metadata" (optional)
        """
        # Separate annotations into batches of 500 for upload
        seperated_images = separate_batches(
            image_batch_size=500, to_batch_list=annotation_list
        )
        for idx, sep_annotations in enumerate(seperated_images, start=1):
            # Prepare annotations for upload
            # Create bulk annotation import job
            annotation_import_job = spb_curate.Annotation.create_bulk(
                dataset_id=self.curate_dataset[
                    "id"
                ],  # Dataset ID to upload annotations to
                annotations=sep_annotations,  # Prepared annotations for upload
            )

            while True:
                # Fetch the current status of the annotation import job
                annotation_import_job = spb_curate.Job.fetch(
                    id=annotation_import_job["id"]
                )

                print(
                    f"[INFO] {(idx-1) * 500 + annotation_import_job['progress']} / {len(annotation_list)} labels updated"
                )

                if annotation_import_job["status"] == "COMPLETE":
                    # Check for and print any failure details if present
                    if annotation_import_job["result"]["fail_detail"]:
                        print(
                            "[INFO] Fail detail: ",
                            annotation_import_job["result"]["fail_detail"],
                        )
                        print(
                            "[INFO] Fail reason: ",
                            annotation_import_job["fail_reason"],
                        )
                    break  # Exit the loop if job is complete

                if annotation_import_job["status"] == "FAILED":
                    if annotation_import_job["result"]["fail_detail"]:
                        print(
                            "[INFO] Fail detail: ",
                            annotation_import_job["result"]["fail_detail"],
                        )
                        print(
                            "[INFO] Fail reason: ",
                            annotation_import_job["fail_reason"],
                        )
                    break
                time.sleep(
                    SLEEP_INTERVAL
                )  # Wait before checking the status again

    def get_width_height(self, data_key: str) -> Tuple[int, int]:
        """
        Fetches the width and height of an image using its data key.

        Args:
            data_key (str): The unique identifier for the image.

        Returns:
            Tuple[int, int]: A tuple containing the width and height of the image.
        """
        image = self.dataset.fetch_images(key=data_key)[0]
        meta = image["metadata"]
        width, height = meta["_width"], meta["_height"]
        return width, height

    def download_image(self, data_key: str, download_path: str):
        """
        Downloads an image from the dataset using its data key.
        """
        image_url = self.dataset.fetch_images(
            key=data_key, include_image_url=True
        )[0]["image_url"]
        try:
            response = requests.get(image_url)
            response.raise_for_status()  # Raises an HTTPError if the response status code is 4XX/5XX
            with open(download_path, "wb") as f:
                f.write(response.content)
            print(f"[INFO] Image downloaded successfully: {download_path}")
        except requests.exceptions.RequestException as e:
            print(f"[ERROR] Failed to download image: {e}")

    def bbox_setting(
        self, data_key: str, class_name: str, annotation: list
    ) -> spb_curate.Annotation:
        """
        Creates a bounding box annotation for a given image.

        Args:
            data_key (str): The unique identifier for the image.
            class_name (str): The class name associated with the bounding box.
            annotation (list): A list containing the x, y coordinates, width, and height of the bounding box.

        Returns:
            spb_curate.Annotation: An Annotation object representing the bounding box.
        """
        bounding_box = spb_curate.BoundingBox(
            raw_data={
                "x": annotation[0],
                "y": annotation[1],
                "width": annotation[2],
                "height": annotation[3],
            },
        )
        bbox_annotation = spb_curate.Annotation(
            image_key=data_key,
            annotation_class=class_name,
            annotation_type="box",
            annotation_value=bounding_box,
            metadata={},
        )

        return bbox_annotation
